"""Processing functions for margarine package."""

import random

import tensorflow as tf
from tensorflow_probability import distributions as tfd


@tf.function(jit_compile=True)
def _forward_transform(
    x: tf.Tensor,
    min: float | int | tf.Tensor = 0,
    max: float | int | tf.Tensor = 1,
) -> tf.Tensor:
    r"""Forward tranforms input samples.

    Normalise between 0 and 1 and then tranform
    onto samples of standard normal distribution (i.e. base of
    tfd.TransformedDistribution).

    Args:
        x: (tf.Tensor) Samples to be normalised.
        min: (float | int | tf.Tensor) Passed from the bijectors code.
        max: (float | int | tf.Tensor) Passed from the bijectors code.
    """
    x = tfd.Uniform(min, max).cdf(x)
    x = tfd.Normal(0, 1).quantile(x)
    return x


@tf.function(jit_compile=True)
def _inverse_transform(
    x: tf.Tensor, min: float | int | tf.Tensor, max: float | int | tf.Tensor
) -> tf.Tensor:
    r"""Inverse tranforms output samples.

    Inverts the processes in ``forward_transofrm``.

    Args:
        x: (tf.Tensor) Samples to be inverse normalised.
        min: (float | int | tf.Tensor) Passed from the bijectors code.
        max: (float | int | tf.Tensor) Passed from the bijectors code.
    """
    x = tfd.Normal(0, 1).cdf(x)
    x = tfd.Uniform(min, max).quantile(x)
    return x


def pure_tf_train_test_split(
    a: tf.Tensor, b: tf.Tensor, test_size: float = 0.2
) -> tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    """Splitting data into training and testing sets.

    Function is equivalent
    to sklearn.model_selection.train_test_split but a and b
    are tensorflow tensors.

    Args:
        a: (tf.Tensor) First set of data to be split.
        b: (tf.Tensor) Second set of data to be split.
        test_size: (float) Proportion of data to be used for testing.
    """
    idx = random.sample(range(len(a)), int(len(a) * test_size))

    a_train = tf.gather(
        a, tf.convert_to_tensor(list(set(range(len(a))) - set(idx)))
    )
    b_train = tf.gather(
        b, tf.convert_to_tensor(list(set(range(len(b))) - set(idx)))
    )
    a_test = tf.gather(a, tf.convert_to_tensor(idx))
    b_test = tf.gather(b, tf.convert_to_tensor(idx))

    return a_train, a_test, b_train, b_test



"""Tests for marginal statistics calculation using MAF and KDE methods."""

import numpy as np
import pytest
import tensorflow as tf
from anesthetic import MCMCSamples
from numpy.testing import assert_allclose, assert_equal
from scipy.stats import norm

from margarine.clustered import clusterMAF
from margarine.kde import KDE
from margarine.maf import MAF
from margarine.marginal_stats import calculate


def D_KL(
    logL: np.ndarray | tf.Tensor, weights: np.ndarray | tf.Tensor
) -> float:
    """Calculate the Kullback-Leibler divergence."""
    return -np.average(logL, weights=weights)


def d_g(
    logL: np.ndarray | tf.Tensor, weights: np.ndarray | tf.Tensor
) -> float:
    """Calculate the BMD statistic."""
    return 2 * np.cov(logL, aweights=weights)


norm = norm(loc=4.2, scale=0.3)
theta = norm.rvs(size=(1000, 2))
logL = norm.logpdf(theta).sum(axis=1)
weights = np.ones(len(theta))

mcmc_samples = MCMCSamples(data=theta, logL=logL)
samples_kl = D_KL(logL, weights)
samples_d = d_g(logL, weights)
names = [i for i in range(theta.shape[-1])]


def test_maf() -> None:
    """Test MAF marginal statistics calculation."""

    def check_stats(label: str) -> None:
        """Check calculated statistics against expected values."""
        if label == "KL Divergence":
            value = samples_kl
            assert_allclose(stats[label], value, rtol=1, atol=1)
        else:
            value = samples_d
            assert_allclose(stats[label], value, rtol=1, atol=1)

    bij = MAF(theta, weights=weights)
    bij.train(10000, early_stop=True, patience=400)

    stats_label = ["KL Divergence", "BMD"]

    stats = calculate(bij).statistics()
    [check_stats(stat_label) for stat_label in stats_label]


def test_maf_kwargs() -> None:
    """Test MAF with incorrect keyword arguments."""
    with pytest.raises(TypeError):
        bij = MAF(theta, weights=weights)
        bij.sample(4.5)
    with pytest.raises(TypeError):
        MAF(theta, weights=weights, number_networks=4.4)
    with pytest.raises(TypeError):
        MAF(theta, weights=weights, learning_rate="foobar")
    with pytest.raises(TypeError):
        MAF(theta, weights=weights, hidden_layers="foobar")
    with pytest.raises(TypeError):
        MAF(theta, weights=weights, hidden_layers=[4.5, 50])
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=4.5)
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=100, early_stop="foo")
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=100, clustering=5)
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=100, cluster_numeber="foo")
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=100, cluster_labels=5)
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=100, patience="foo")


def test_maf_save_load() -> None:
    """Test the save and load functionality of MAF."""
    bij = MAF(theta, weights=weights)
    bij.train(100)
    file = "saved_maf.pkl"
    bij.save(file)
    loaded_bijector = MAF.load(file)
    for i in range(len(bij.mades)):
        assert_equal(
            bij.mades[i].get_weights(), loaded_bijector.mades[i].get_weights()
        )


def test_kde() -> None:
    """Test KDE marginal statistics calculation."""

    def check_stats(label: str) -> None:
        """Check calculated statistics against expected values."""
        if label == "KL Divergence":
            value = samples_kl
            assert_allclose(stats[label], value, rtol=1, atol=1)
        else:
            value = samples_d
            assert_allclose(stats[label], value, rtol=1, atol=1)

    kde = KDE(theta, weights=weights)
    kde.generate_kde()

    stats_label = ["KL Divergence", "BMD"]

    stats = calculate(kde).statistics()
    [check_stats(stat_label) for stat_label in stats_label]


def test_kde_save_load() -> None:
    """Test the save and load functionality of KDE."""
    kde = KDE(theta, weights=weights)
    kde.generate_kde()
    file = "saved_maf.pkl"
    kde.save(file)
    loaded_kde = KDE.load(file)
    assert_equal(kde.kde.covariance, loaded_kde.kde.covariance)


def test_anesthetic() -> None:
    """Test MAF and KDE with anesthetic MCMCSamples object."""
    kde = KDE(mcmc_samples, parameters=names)
    maf = MAF(mcmc_samples, parameters=names)
    cmaf = clusterMAF(mcmc_samples)

    assert_equal(kde.parameters, names)
    assert_equal(maf.parameters, names)

    # not providing parametes here but deriving them from the
    # anesthetic object columns
    assert np.all(cmaf.parameters == np.array(names))



"""Module for calculating marginal statistics from density estimators."""

from collections.abc import Callable

import numpy as np
from scipy.special import logsumexp
from tensorflow_probability import distributions as tfd
from tqdm import tqdm

from margarine.clustered import clusterMAF
from margarine.kde import KDE
from margarine.maf import MAF


class calculate:
    """Class to calculate marginal statistics from density estimators."""

    def __init__(
        self,
        de: MAF | KDE | clusterMAF,
        **kwargs: dict,
    ) -> None:
        r"""Calculate marginal KL divergences and Bayesian dimensionalities.

        This class uses a trained MAF or KDE along with
        generated samples to compute
        information-theoretic measures of the posterior distribution.

        Args:
            de (MAF | KDE | clusterMAF): A trained and
                loaded instance of MAF,
                clusterMAF, or KDE.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            prior_de (MAF | clusterMAF | KDE, optional): A trained and loaded
                instance of MAF, clusterMAF, or KDE representing the prior
                distribution. If not provided, a uniform prior will be used.

        Example:
            >>> from margarine.maf import MAF
            >>> from margarine.kde import KDE
            >>> from margarine.clustered import clusterMAF
            >>> import numpy as np
            >>>
            >>> # Load a trained model
            >>> maf = MAF.load('/trained_maf.pkl')
            >>> # or
            >>> kde = KDE.load('/trained_kde.pkl')
            >>> # or
            >>> clustered_maf = clusterMAF.load('/trained_clustered_maf.pkl')
            >>>
            >>> # Generate samples
            >>> u = np.random.uniform(0, 1, size=(10000, 5))
            >>> prior_limits = np.array([[0]*5, [1]*5])
            >>> samples = maf(u, prior_limits)
            >>>
            >>> # Initialize the class
            >>> stats = calculate(maf, samples)
        """
        self.de = de

        self.theta = self.de.theta
        self.theta_weights = self.de.sample_weights
        self.theta_min = self.de.theta_min
        self.theta_max = self.de.theta_max

        if isinstance(self.de, MAF):
            self.theta_weights = self.theta_weights.numpy()
            self.theta_min = self.theta_min.numpy()
            self.theta_max = self.theta_max.numpy()

        self.samples = self.de.sample(len(self.theta))

        self.prior_de = kwargs.pop("prior_de", None)

    def statistics(self) -> dict:
        """Calculate marginal statistics.

        Calculate the marginal bayesian KL divergence and dimensionality with
        approximate errors.

        Returns:
            results (dict): Dictionary containing the calculated statistics.
        """

        def mask_arr(arr: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
            return arr[np.isfinite(arr)], np.isfinite(arr)

        logprob = self.de.log_prob(self.samples)
        theta_logprob = self.de.log_prob(self.theta)

        if isinstance(self.de, MAF):
            logprob = logprob.numpy()
            theta_logprob = theta_logprob.numpy()

        args = np.argsort(theta_logprob)
        self.theta_weights = self.theta_weights[args]
        theta_logprob = theta_logprob[args]

        deargs = np.argsort(logprob)
        logprob = logprob[deargs]
        wde = [np.sum(self.theta_weights) / len(logprob)] * len(logprob)
        logprob = np.interp(
            np.cumsum(self.theta_weights), np.cumsum(wde), logprob
        )

        mid_point = np.log((np.exp(logprob) + np.exp(theta_logprob)) / 2)

        if self.prior_de is None:
            self.base = tfd.Blockwise(
                [
                    tfd.Uniform(self.theta_min[i], self.theta_max[i])
                    for i in range(self.samples.shape[-1])
                ]
            )
            prior = self.base.sample(len(self.theta)).numpy()

            theta_base_logprob = self.base.log_prob(self.theta).numpy()
            base_logprob = self.base.log_prob(prior).numpy()

            prior_wde = [np.sum(self.theta_weights) / len(base_logprob)] * len(
                base_logprob
            )

            base_logprob = base_logprob[deargs]
            base_logprob = np.interp(
                np.cumsum(self.theta_weights),
                np.cumsum(prior_wde),
                base_logprob,
            )

            theta_base_logprob = theta_base_logprob[args]
        elif self.prior_de is not None:
            self.base = self.prior_de
            de_prior_samples = self.base.sample(len(self.theta))
            theta_base_logprob = self.base.log_prob(de_prior_samples)
            base_logprob = self.base.log_prob(self.theta)

            if isinstance(self.prior_de, MAF):
                theta_base_logprob = theta_base_logprob.numpy()
                base_logprob = base_logprob.numpy()

            base_logprob = base_logprob[deargs]
            base_logprob = np.interp(
                np.cumsum(self.theta_weights), np.cumsum(wde), base_logprob
            )

            base_args = np.argsort(theta_base_logprob)
            theta_base_logprob = theta_base_logprob[base_args]
            prior_weights = np.cumsum(
                self.prior_de.sample_weights.numpy()[base_args]
            )

            prior_weights = (
                np.cumsum(self.theta_weights).max()
                - np.cumsum(self.theta_weights).min()
            ) * (prior_weights - prior_weights.min()) / (
                prior_weights.max() - prior_weights.min()
            ) + np.cumsum(self.theta_weights).min()

            theta_base_logprob = np.interp(
                np.cumsum(self.theta_weights),
                prior_weights,
                theta_base_logprob,
            )

        midbasepoint = np.log(
            (np.exp(base_logprob) + np.exp(theta_base_logprob)) / 2
        )

        mid_logL, mask = mask_arr(mid_point - midbasepoint)
        midkl = np.average(mid_logL, weights=self.theta_weights[mask])
        midbd = 2 * np.cov(mid_logL, aweights=self.theta_weights[mask])
        de_logL, mask = mask_arr(logprob - base_logprob)
        dekl = np.average(de_logL, weights=self.theta_weights[mask])
        debd = 2 * np.cov(de_logL, aweights=self.theta_weights[mask])
        theta_logL, mask = mask_arr(theta_logprob - theta_base_logprob)
        thetakl = np.average(theta_logL, weights=self.theta_weights[mask])
        thetabd = 2 * np.cov(theta_logL, aweights=self.theta_weights[mask])

        kl_array = np.sort([midkl, dekl, thetakl])
        bd_array = np.sort([midbd, debd, thetabd])

        results = {
            "KL Divergence": kl_array[1],
            "KL Lower Bound": kl_array[0],
            "KL Upper Bound": kl_array[2],
            "BMD": bd_array[1],
            "BMD Lower Bound": bd_array[0],
            "BMD Upper Bound": bd_array[2],
        }

        return results

    def integrate(
        self,
        loglikelihood: Callable,
        prior_pdf: Callable,
        batch_size: int = 1000,
        sample_size: int = 10000,
        logzero: float = -1e30,
    ) -> dict:
        """Importance sampling integration of a likelihood function.

        Args:
            loglikelihood (Callable): A function that takes a
                    numpy array of samples
                    and returns the loglikelihood of each sample.
            prior_pdf (Callable): A function that takes a numpy
                    array of samples and returns the prior
                    logpdf of each sample.
            batch_size (int): The number of samples to draw at each iteration.
            sample_size (int): The number of samples to draw in total.
            logzero (float): The definition of zero for the
                loglikelihood function.

        Returns:
            stats (dict): Dictionary containing useful statistics
        """
        xs = np.empty((sample_size, self.de.theta.shape[-1]))
        fs = np.empty(sample_size)
        gs = np.empty(sample_size)
        pis = np.empty(sample_size)

        n_todo = sample_size
        trials = 0

        with tqdm(total=sample_size) as pbar:
            while n_todo > 0:
                x = self.de.sample(batch_size).numpy()
                f = np.array(list(map(loglikelihood, x)))
                g = self.de.log_prob(x).numpy()
                in_bounds = np.logical_and(f >= logzero, g >= logzero)
                n_accept = x[in_bounds].shape[0]
                if n_accept <= n_todo:
                    xs[
                        sample_size - n_todo : sample_size - n_todo + n_accept
                    ] = x[in_bounds]
                    fs[
                        sample_size - n_todo : sample_size - n_todo + n_accept
                    ] = f[in_bounds]
                    gs[
                        sample_size - n_todo : sample_size - n_todo + n_accept
                    ] = g[in_bounds]
                    pis[
                        sample_size - n_todo : sample_size - n_todo + n_accept
                    ] = prior_pdf(x[in_bounds])
                    trials += batch_size
                else:
                    n_accept = n_todo
                    xs[sample_size - n_todo :] = x[in_bounds][:n_accept]
                    fs[sample_size - n_todo :] = f[in_bounds][:n_accept]
                    gs[sample_size - n_todo :] = g[in_bounds][:n_accept]
                    pis[sample_size - n_todo :] = prior_pdf(x[in_bounds])[
                        :n_accept
                    ]
                    last_index = in_bounds[-1]
                    trials += last_index + 1
                n_todo -= n_accept
                pbar.update(n_accept)
                if trials > 10 * sample_size:
                    raise ValueError(
                        "Too many unsuccessful trials, this"
                        + "typically indicates mismatch between"
                        + "flow and likelihood"
                    )

            weights = np.exp(fs + pis - gs)

            eff = np.sum(weights) ** 2 / np.sum(weights**2) / sample_size
            integral = sample_size / trials * weights.mean()
            log_integral = logsumexp(fs + pis - gs) - np.log(trials)

            stderr = np.sqrt(
                (np.sum(weights**2) / trials - integral**2) / (trials - 1)
            )
            log_stderr = np.log(stderr)

        stats = {
            "x": xs,
            "y": fs,
            "weights": weights,
            "efficiency": eff,
            "trials": trials,
            "log_integral": log_integral,
            "log_stderr": log_stderr,
        }
        return stats



"""This module contains the MAF class used to build, train and call."""

import pickle
import warnings

import anesthetic
import numpy as np
import tensorflow as tf
import tqdm
from anesthetic.samples import Samples
from tensorflow_probability import bijectors as tfb
from tensorflow_probability import distributions as tfd

from margarine.processing import (
    _forward_transform,
    _inverse_transform,
    pure_tf_train_test_split,
)


class MAF:
    r"""The MAF class.

    This class is used to train, load and call instances of a bijector
    built from a series of autoregressive neural networks.
    """

    def __init__(
        self,
        theta: tf.Tensor | np.ndarray | Samples,
        **kwargs: dict,
    ) -> None:
        """Initialize the MAF class.

        Args:
            theta (tf.Tensor | np.ndarray | Samples): The
                samples from the
                probability distribution that the MAF should learn.
                Can be a numpy array or an
                anesthetic NestedSamples or MCMCSamples object.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            weights (np.ndarray, optional): The weights associated with the
                samples. If an anesthetic NestedSamples or MCMCSamples object
                is passed, the weights are drawn from it.
                Defaults to np.ones(len(theta)).
            number_networks (int, optional): The number of
                autoregressive neural networks to chain together
                in the bijector. Defaults to 6.
            learning_rate (float, optional): The step size of the optimization
                algorithm used to train the MAF. Can affect the
                quality of emulation.
                Defaults to 1e-3.
            hidden_layers (list, optional): The number of layers
                and nodes in each hidden layer for each neural network.
                Each network in the chain has the same hidden layer structure.
                Defaults to [50, 50].
            activation_func (str, optional): The activation function keyword
                recognizable by TensorFlow. Defaults to 'tanh'.
            theta_max (np.ndarray, optional): The true upper
                limits of the priors used to generate the samples.
            theta_min (np.ndarray, optional): The true lower limits
                of the priors used to generate the samples.
            parameters (list of str, optional): The relevant
                parameters to train on. Only needed if theta is an
                anesthetic samples object. If not provided,
                all parameters will be used.

        Attributes:
            theta_max (np.ndarray): The true upper limits of the priors. If not
                supplied as a keyword argument, this is an approximate
                estimate.
            theta_min (np.ndarray): The true lower limits of the priors. If not
                supplied as a keyword argument, this is an approximate
                estimate.
            loss_history (list): The value of the loss function
                at each epoch during training.
        """
        self.number_networks = kwargs.pop("number_networks", 6)
        self.learning_rate = kwargs.pop("learning_rate", 1e-3)
        self.hidden_layers = kwargs.pop("hidden_layers", [50, 50])
        self.parameters = kwargs.pop("parameters", None)
        self.activation_func = kwargs.pop("activation_func", "tanh")

        # Avoids unintended side effects outside the class
        if not isinstance(theta, tf.Tensor):
            theta = theta.copy()
        else:
            theta = tf.identity(theta)

        if isinstance(
            theta,
            anesthetic.samples.NestedSamples | anesthetic.samples.MCMCSamples,
        ):
            weights = theta.get_weights()
            if self.parameters:
                theta = theta[self.parameters].values
            else:
                if isinstance(theta, anesthetic.samples.NestedSamples):
                    self.parameters = theta.columns[:-3].values
                    theta = theta[theta.columns[:-3]].values
                else:
                    self.parameters = theta.columns[:-1].values
                    theta = theta[theta.columns[:-1]].values
        else:
            weights = kwargs.pop("weights", np.ones(len(theta)))

        self.theta = tf.convert_to_tensor(theta, dtype=tf.float32)
        if not isinstance(weights, tf.Tensor):
            weights = tf.convert_to_tensor(weights.copy(), dtype=tf.float32)
        else:
            weights = tf.identity(weights)
        if weights.dtype != tf.float32:
            weights = tf.cast(weights, tf.float32)
        self.sample_weights = weights

        mask = np.isfinite(theta).all(axis=-1)
        self.theta = tf.boolean_mask(self.theta, mask, axis=0)
        self.sample_weights = tf.boolean_mask(
            self.sample_weights, mask, axis=0
        )

        self.n = tf.math.reduce_sum(
            self.sample_weights
        ) ** 2 / tf.math.reduce_sum(self.sample_weights**2)

        theta_max = tf.math.reduce_max(self.theta, axis=0)
        theta_min = tf.math.reduce_min(self.theta, axis=0)
        a = ((self.n - 2) * theta_max - theta_min) / (self.n - 3)
        b = ((self.n - 2) * theta_min - theta_max) / (self.n - 3)
        self.theta_min = kwargs.pop("theta_min", b)
        self.theta_max = kwargs.pop("theta_max", a)

        # Convert to tensors if not already
        if not isinstance(self.theta_min, tf.Tensor):
            self.theta_min = tf.convert_to_tensor(
                self.theta_min.copy(), dtype=tf.float32
            )
        if not isinstance(self.theta_max, tf.Tensor):
            self.theta_max = tf.convert_to_tensor(
                self.theta_max.copy(), dtype=tf.float32
            )

        # Discard samples outside the prior range, provide a warning if any
        # are discarded
        mask = tf.math.reduce_all(
            (self.theta >= self.theta_min) & (self.theta <= self.theta_max),
            axis=-1,
        )
        if not tf.math.reduce_all(mask):
            warnings.warn(
                "Some samples are outside the user specified prior range "
                "and will be discarded! The specified range is likely smaller "
                "than the range covered by the samples."
            )
            self.theta = tf.boolean_mask(self.theta, mask, axis=0)
            self.sample_weights = tf.boolean_mask(
                self.sample_weights, mask, axis=0
            )

        if type(self.number_networks) is not int:
            raise TypeError("'number_networks' must be an integer.")
        if not isinstance(
            self.learning_rate,
            int | float | tf.keras.optimizers.schedules.LearningRateSchedule,
        ):
            raise TypeError(
                "'learning_rate', "
                + "must be an integer, float or tf.keras scheduler."
            )
        if type(self.hidden_layers) is not list:
            raise TypeError("'hidden_layers' must be a list of integers.")
        else:
            for i in range(len(self.hidden_layers)):
                if type(self.hidden_layers[i]) is not int:
                    raise TypeError(
                        "One or more values in 'hidden_layers'"
                        + "is not an integer."
                    )

        self.optimizer = tf.keras.optimizers.Adam(
            learning_rate=self.learning_rate
        )

        self.gen_mades()

    def gen_mades(self) -> tuple[tfb.Bijector, tfd.TransformedDistribution]:
        """Generating the masked autoregressive flow."""
        self.mades = [
            tfb.AutoregressiveNetwork(
                params=2,
                hidden_units=self.hidden_layers,
                activation=self.activation_func,
                input_order="random",
            )
            for _ in range(self.number_networks)
        ]

        self.bij = tfb.Chain(
            [tfb.MaskedAutoregressiveFlow(made) for made in self.mades]
        )

        self.base = tfd.Blockwise(
            [tfd.Normal(loc=0, scale=1) for _ in range(len(self.theta_min))]
        )

        self.maf = tfd.TransformedDistribution(self.base, bijector=self.bij)

        return self.bij, self.maf

    def train(
        self,
        epochs: int = 100,
        patience: int | None = None,
        early_stop: bool = False,
        loss_type: str = "sum",
    ) -> None:
        r"""Train the MAF once it has been initialized.

        This method calls the internal `_training()` function
        to perform the training process.

        Example:
            >>> from margarine.maf import MAF
            >>> bij = MAF(theta, weights=weights)
            >>> bij.train()

        Keyword Args:
            epochs (int, optional): The number of iterations to
                train the neural networks for. Defaults to 100.
            patience (int | None, optional): The number of epochs with
                no improvement on the test loss before early
                stopping is triggered.
                Defaults to 2% of total requested epochs.
            early_stop (bool, optional): Whether to implement
                early stopping or train for the set number of epochs.
                If True, training stops when test loss
                has not improved for 2% of the requested epochs,
                and the model rolls back to the best performing version.
                Defaults to False.
            loss_type (str, optional): Whether to use the 'sum' or
                'mean' of the weighted log probabilities to
                calculate the loss function.
                Defaults to 'sum'.
        """
        if type(epochs) is not int:
            raise TypeError("'epochs' is not an integer.")
        if type(early_stop) is not bool:
            raise TypeError("'early_stop' must be a boolean.")
        if patience is not None and type(patience) is not int:
            raise TypeError("'patience' must be an integer or None.")

        self.epochs = epochs
        self.early_stop = early_stop
        self.loss_type = loss_type

        if patience is None:
            self.patience = round((self.epochs / 100) * 2)
        else:
            self.patience = patience

        self.maf = self._training(
            self.theta,
            self.patience,
            self.sample_weights,
            self.maf,
            self.theta_min,
            self.theta_max,
        )

    def _training(
        self,
        theta: tf.Tensor,
        patience: int,
        sample_weights: tf.Tensor,
        maf: tfd.TransformedDistribution,
        theta_min: float | tf.Tensor,
        theta_max: float | tf.Tensor,
    ) -> tfd.TransformedDistribution:
        """Training the masked autoregressive flow.

        This function performs the training of the MAF by splitting
        the data into training and testing sets, then iteratively
        updating the weights of the neural networks based on the
        calculated loss.

        Args:
            theta (tf.Tensor): The samples to train on.
            patience (int): The number of epochs with no improvement
                on the test loss before early stopping is triggered.
            sample_weights (tf.Tensor): The weights associated with the
                samples.
            maf (tfd.TransformedDistribution): The MAF to be trained.
            theta_min (float | tf.Tensor): The minimum values of the priors.
            theta_max (float | tf.Tensor): The maximum values of the priors.

        Returns:
            tfd.TransformedDistribution: The trained MAF.
        """
        phi = _forward_transform(theta, theta_min, theta_max)
        weights_phi = sample_weights / tf.reduce_sum(sample_weights)

        phi_train, phi_test, weights_phi_train, weights_phi_test = (
            pure_tf_train_test_split(phi, weights_phi, test_size=0.2)
        )

        self.loss_history = []
        self.test_loss_history = []
        c = 0
        for i in tqdm.tqdm(range(self.epochs)):
            loss = self._train_step(
                phi_train, weights_phi_train, self.loss_type, maf
            )
            self.loss_history.append(loss)

            self.test_loss_history.append(
                self._test_step(
                    phi_test, weights_phi_test, self.loss_type, maf
                )
            )

            if self.early_stop:
                c += 1
                if i == 0:
                    minimum_loss = self.test_loss_history[-1]
                    minimum_epoch = i
                    minimum_model = None
                else:
                    if self.test_loss_history[-1] < minimum_loss:
                        minimum_loss = self.test_loss_history[-1]
                        minimum_epoch = i
                        minimum_model = maf.copy()
                        c = 0
                if minimum_model:
                    if c == patience:
                        print(
                            "Early stopped. Epochs used = "
                            + str(i)
                            + ". Minimum at epoch = "
                            + str(minimum_epoch)
                        )
                        return minimum_model
        return maf

    @tf.function(jit_compile=True)
    def _test_step(
        self,
        x: tf.Tensor,
        w: tf.Tensor,
        loss_type: str,
        maf: tfd.TransformedDistribution,
    ) -> tf.Tensor:
        r"""Calculate the test loss.

        This function is used to calculate the test loss value at each epoch
        for early stopping.

        Args:
            x (tf.Tensor): The test samples.
            w (tf.Tensor): The weights associated with the test samples.
            loss_type (str): The type of loss to calculate ('sum' or 'mean').
            maf (tfd.TransformedDistribution): The MAF being trained.

        Returns:
            tf.Tensor: The calculated loss value.
        """
        if loss_type == "sum":
            loss = -tf.reduce_sum(w * maf.log_prob(x))
        elif loss_type == "mean":
            loss = -tf.reduce_mean(w * maf.log_prob(x))
        return loss

    @tf.function(jit_compile=True)
    def _train_step(
        self,
        x: tf.Tensor,
        w: tf.Tensor,
        loss_type: str,
        maf: tfd.TransformedDistribution,
    ) -> tf.Tensor:
        r"""Calculate the training loss and apply gradients.

        This function is used to calculate the loss value at each epoch and
        adjust the weights and biases of the neural networks via the
        optimizer algorithm.

        Args:
            x (tf.Tensor): The training samples.
            w (tf.Tensor): The weights associated with the training samples.
            loss_type (str): The type of loss to calculate ('sum' or 'mean').
            maf (tfd.TransformedDistribution): The MAF being trained.

        Returns:
            tf.Tensor: The calculated loss value.
        """
        with tf.GradientTape() as tape:
            if loss_type == "sum":
                loss = -tf.reduce_sum(w * maf.log_prob(x))
            elif loss_type == "mean":
                loss = -tf.reduce_mean(w * maf.log_prob(x))
        gradients = tape.gradient(loss, maf.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, maf.trainable_variables))
        return loss

    @tf.function(jit_compile=True)
    def __call__(self, u: tf.Tensor | np.ndarray) -> tf.Tensor:
        r"""Transform samples from the unit hypercube to samples on the MAF.

        Args:
            u (tf.Tensor | np.ndarray): Samples from the unit hypercube
                to be transformed.

        Returns:
            tf.Tensor: The transformed samples.
        """
        if u.dtype != tf.float32:
            u = tf.cast(u, tf.float32)

        x = _forward_transform(u)
        x = self.bij(x)
        x = _inverse_transform(x, self.theta_min, self.theta_max)

        return x

    @tf.function(jit_compile=True)
    def sample(self, length: int = 1000) -> tf.Tensor:
        r"""Generate samples from the MAF.

        Args:
            length (int, optional): The number of samples to generate.
                Defaults to 1000.

        Returns:
            tf.Tensor: The generated samples.
        """
        if type(length) is not int:
            raise TypeError("'length' must be an integer.")

        u = tf.random.uniform((length, len(self.theta_min)))
        return self(u)

    @tf.function(jit_compile=True, reduce_retracing=True)
    def log_prob(self, params: tf.Tensor | np.ndarray) -> tf.Tensor:
        """Caluclate the log-probability for a given set of parameters.

        While the density estimator has its own built in log probability
        function, a correction has to be applied for the transformation of
        variables that is used to improve accuracy when learning. The
        correction is implemented here.

        Args:
            params (tf.Tensor | np.ndarray): The set of samples for which to
                calculate the log probability.

        Returns:
            tf.Tensor: The log-probabilities of the provided samples.

        """
        # Enforce float32 dtype
        if params.dtype != tf.float32:
            params = tf.cast(params, tf.float32)

        def calc_log_prob(
            mins: float | int | tf.Tensor,
            maxs: float | int | tf.Tensor,
            maf: tfd.TransformedDistribution,
        ) -> tf.Tensor:
            """Function to calculate log-probability for a given MAF.

            Args:
                mins (float | int | tf.Tensor): Minimum prior values.
                maxs (float | int | tf.Tensor): Maximum prior values.
                maf (tfd.TransformedDistribution): The MAF to calculate
                    the log-probability for.

            Returns:
                tf.Tensor: The log-probabilities of the provided samples.
            """

            def norm_jac(y: tf.Tensor) -> tf.Tensor:
                return transform_chain.inverse_log_det_jacobian(
                    y, event_ndims=0
                )

            transformed_x = _forward_transform(params, mins, maxs)

            transform_chain = tfb.Chain(
                [
                    tfb.Invert(tfb.NormalCDF()),
                    tfb.Scale(1 / (maxs - mins)),
                    tfb.Shift(-mins),
                ]
            )

            correction = norm_jac(transformed_x)
            logprob = maf.log_prob(transformed_x) - tf.reduce_sum(
                correction, axis=-1
            )
            return logprob

        logprob = calc_log_prob(self.theta_min, self.theta_max, self.maf)

        return logprob

    def log_like(
        self,
        params: tf.Tensor | np.ndarray,
        logevidence: float,
        prior_de: tfd.TransformedDistribution = None,
    ) -> tf.Tensor:
        r"""Return the log-likelihood for a given set of parameters.

        It requires the logevidence from the original nested sampling run
        in order to do this and in the case that the prior is non-uniform
        a trained prior density estimator should be provided.

        Args:
            params (tf.Tensor | np.ndarray): The set of samples for which to
                calculate the log-likelihood.
            logevidence (float): The log-evidence from the original
                nested sampling run.
            prior_de (tfd.TransformedDistribution, optional): A trained
                density estimator for the prior. If the prior is uniform,
                this can be left as None. Defaults to None.

        Returns:
            tf.Tensor: The log-likelihoods of the provided samples.
        """
        if prior_de is None:
            warnings.warn("Assuming prior is uniform!")
            prior_logprob = tf.math.log(
                tf.math.reduce_prod(
                    [
                        1 / (self.theta_max[i] - self.theta_min[i])
                        for i in range(len(self.theta_min))
                    ]
                )
            )
        else:
            prior_logprob = self.prior_de.log_prob(params)

        posterior_logprob = self.log_prob(params)

        loglike = posterior_logprob + logevidence - prior_logprob

        return loglike

    def save(self, filename: str) -> None:
        r"""Save an instance of a trained MAF.

        Args:
            filename (str): Path to save the MAF to.
        """
        nn_weights = [made.get_weights() for made in self.mades]
        with open(filename, "wb") as f:
            pickle.dump(
                [
                    self.theta,
                    nn_weights,
                    self.sample_weights,
                    self.number_networks,
                    self.hidden_layers,
                    self.learning_rate,
                    self.theta_min,
                    self.theta_max,
                ],
                f,
            )

    @classmethod
    def load(cls, filename: str) -> "MAF":
        r"""Load a saved MAF from a file.

        Example:
            >>> from margarine.maf import MAF
            >>> file = 'path/to/pickled/MAF.pkl'
            >>> bij = MAF.load(file)

        Args:
            filename (str): Path to the saved MAF file.

        Returns:
            MAF: The loaded MAF object.
        """
        with open(filename, "rb") as f:
            data = pickle.load(f)
            (
                theta,
                nn_weights,
                sample_weights,
                number_networks,
                hidden_layers,
                learning_rate,
                theta_min,
                theta_max,
            ) = data

        bijector = cls(
            theta,
            weights=sample_weights,
            number_networks=number_networks,
            learning_rate=learning_rate,
            hidden_layers=hidden_layers,
            theta_min=theta_min,
            theta_max=theta_max,
        )
        bijector(np.random.uniform(0, 1, size=(len(theta), theta.shape[-1])))
        for made, nn_weights in zip(bijector.mades, nn_weights):
            made.set_weights(nn_weights)

        return bijector



"""Tests for clustered MAF flows and their marginal statistics."""

import numpy as np
import pytest
import tensorflow as tf
from anesthetic import MCMCSamples
from clustered_distributions import TwoMoons
from numpy.testing import assert_allclose, assert_equal

from margarine.clustered import clusterMAF
from margarine.marginal_stats import calculate


def D_KL(
    logL: np.ndarray | tf.Tensor, weights: np.ndarray | tf.Tensor
) -> float:
    """Calculate the Kullback-Leibler divergence."""
    return -np.average(logL, weights=weights)


def d_g(
    logL: np.ndarray | tf.Tensor, weights: np.ndarray | tf.Tensor
) -> float:
    """Calculate the BMD statistic."""
    return 2 * np.cov(logL, aweights=weights)


nsamples = 2500

tm = TwoMoons()

samples = tm.sample(nsamples)
theta = samples.numpy()
weights = np.ones(len(theta))

logL = tm.log_prob(samples).numpy()

samples = MCMCSamples(data=theta)
samples_kl = D_KL(logL, weights)
samples_d = d_g(logL, weights)
names = [i for i in range(theta.shape[-1])]


def test_maf_clustering() -> None:
    """Test clustered MAF marginal statistics calculation."""
    bij = clusterMAF(theta, weights=weights)
    bij.train(10000, early_stop=True, patience=400)
    file = "saved_maf_cluster.pkl"
    bij.save(file)

    loaded_bijector = clusterMAF.load("saved_maf_cluster.pkl")
    for f in range(len(bij.flow)):
        for i in range(len(bij.flow[f].mades)):
            assert_equal(
                bij.flow[f].mades[i].get_weights(),
                loaded_bijector.flow[f].mades[i].get_weights(),
            )

    def check_stats(label: str) -> None:
        """Check calculated statistics against expected values."""
        if label == "KL Divergence":
            value = samples_kl
            assert_allclose(stats[label], value, rtol=1, atol=1)
        else:
            value = samples_d
            assert_allclose(stats[label], value, rtol=1, atol=1)

    stats_label = ["KL Divergence", "BMD"]

    stats = calculate(bij).statistics()
    [check_stats(stat_label) for stat_label in stats_label]


def test_maf_cluster_kwargs() -> None:
    """Test clustered MAF with incorrect keyword arguments."""
    with pytest.raises(ValueError):
        clusterMAF(theta, weights=weights, cluster_number=3)
    with pytest.raises(ValueError):
        clusterMAF(theta, weightes=weights, cluster_labels=np.ones(len(theta)))

    labels = np.ones(len(theta))
    labels[: len(theta) // 2] = 0


def test_cluster_size() -> None:
    """Test clustered MAF with small cluster sizes."""

    # testing the while loop for cluster size
    def draw(ndims: int) -> np.ndarray:
        """Draw samples from a multivariate normal distribution."""
        return np.random.multivariate_normal(
            np.zeros(ndims), np.eye(ndims), ndims * 10
        )

    samples = draw(3)
    clusterMAF(samples)



"""Module for clustered mixture of masked autoregressive flows (MAFs)."""

import pickle
import warnings

import anesthetic
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from anesthetic.samples import Samples
from scipy.special import logsumexp
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score

from margarine.maf import MAF

tfd = tfp.distributions


class clusterMAF:
    """Create clustered mixture of MAFs to model multi-modal distributions."""

    def __init__(
        self,
        theta: tf.Tensor | np.ndarray | Samples,
        **kwargs: dict,
    ) -> None:
        r"""Piecewise normalizing flow built from masked autoregressive flows.

        This class is a wrapper around the MAF class with additional clustering
        functionality. It trains, loads, and
        calls a piecewise normalizing flow where
        different MAFs are trained on different clusters of the sample space.
        This class has all the same functionality as the MAF class
        and can be used in the
        same way.

        Args:
            theta (tf.Tensor | np.ndarray | Samples):
                The samples from the probability distribution to learn.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            weights (np.ndarray, optional): The weights associated with
                the samples. If an anesthetic NestedSamples or MCMCSamples
                object is passed, the weights are drawn from it.
                Defaults to np.ones(len(theta)).
            number_networks (int, optional): The number of autoregressive
                neural networks to chain together in each MAF bijector.
                Defaults to 6.
            learning_rate (float, optional): The step size of the optimization
                algorithm used to train each MAF. Can affect the
                quality of emulation.
                Defaults to 1e-3.
            hidden_layers (list, optional): The number of layers
                and nodes in each
                hidden layer for each neural network. Each network
                in the chain has the same hidden layer structure.
                Defaults to [50, 50].
            activation_func (str, optional): The activation function keyword
                recognizable by TensorFlow. Defaults to 'tanh'.
            cluster_labels (list of int, optional): Labels for the samples if
                clustering has been performed externally.
                Labels should be integers
                from 0 to k-1 corresponding to the cluster
                each sample belongs to. Defaults to None.
            cluster_number (int, optional): The number of clusters k,
                required if cluster_labels is provided. Defaults to None.
            parameters (list of str, optional): The relevant
                parameters to train on. Only needed if theta is an
                anesthetic samples object.
                If not provided, all parameters will be used.

        Note:
            If cluster_labels and cluster_number are not provided,
            clustering will be
            performed automatically during training.
        """
        # Unpack kwargs
        self.number_networks = kwargs.pop("number_networks", 6)
        self.learning_rate = kwargs.pop("learning_rate", 1e-3)
        self.hidden_layers = kwargs.pop("hidden_layers", [50, 50])
        self.cluster_labels = kwargs.pop("cluster_labels", None)
        self.cluster_number = kwargs.pop("cluster_number", None)
        self.parameters = kwargs.pop("parameters", None)
        self.activation_func = kwargs.pop("activation_func", "tanh")

        # Avoid unintended side effects by copying theta
        theta = theta.copy()

        if isinstance(
            theta,
            anesthetic.samples.NestedSamples | anesthetic.samples.MCMCSamples,
        ):
            weights = theta.get_weights()
            if self.parameters:
                theta = theta[self.parameters].values
            else:
                if isinstance(theta, anesthetic.samples.NestedSamples):
                    self.parameters = theta.columns[:-3].values
                    theta = theta[theta.columns[:-3]].values
                else:
                    self.parameters = theta.columns[:-1].values
                    theta = theta[theta.columns[:-1]].values
        else:
            weights = kwargs.pop("weights", np.ones(len(theta)))

        self.theta = theta
        self.sample_weights = weights.copy()

        # needed for marginal stats
        self.n = np.sum(self.sample_weights) ** 2 / np.sum(
            self.sample_weights**2
        )
        theta_max = np.max(self.theta, axis=0)
        theta_min = np.min(self.theta, axis=0)
        a = ((self.n - 2) * theta_max - theta_min) / (self.n - 3)
        b = ((self.n - 2) * theta_min - theta_max) / (self.n - 3)
        self.theta_min = kwargs.pop("theta_min", b)
        self.theta_max = kwargs.pop("theta_max", a)

        # Convert min and max to float 32 if needed
        if not isinstance(self.theta_min, tf.Tensor):
            self.theta_min = tf.convert_to_tensor(
                self.theta_min.copy(), dtype=tf.float32
            )
        if not isinstance(self.theta_max, tf.Tensor):
            self.theta_max = tf.convert_to_tensor(
                self.theta_max.copy(), dtype=tf.float32
            )

        if type(self.number_networks) is not int:
            raise TypeError("'number_networks' must be an integer.")
        if not isinstance(
            self.learning_rate,
            int | float | tf.keras.optimizers.schedules.LearningRateSchedule,
        ):
            raise TypeError(
                "'learning_rate', "
                + "must be an integer, float or tf.keras scheduler."
            )
        if type(self.hidden_layers) is not list:
            raise TypeError("'hidden_layers' must be a list of integers.")
        else:
            for i in range(len(self.hidden_layers)):
                if type(self.hidden_layers[i]) is not int:
                    raise TypeError(
                        "One or more values in 'hidden_layers'"
                        + "is not an integer."
                    )

        mask = np.isfinite(theta).all(axis=-1)
        self.theta = self.theta[mask]
        self.sample_weights = self.sample_weights[mask]

        if self.cluster_number is not None:
            if self.cluster_labels is None:
                raise ValueError(
                    "'cluster_labels' should be provided if "
                    + "'cluster_number' is specified."
                )
        else:
            if self.cluster_labels is not None:
                raise ValueError(
                    "'cluster_number' should be provided if "
                    + "'cluster_labels' is specified."
                )

        if self.cluster_number is not None:
            if self.cluster_number % 1 != 0:
                raise TypeError("'cluster_number' must be a whole number.")
        if self.cluster_labels is not None:
            if not isinstance(self.cluster_labels, np.ndarray | list):
                raise TypeError(
                    "'cluster_labels' must be an array " + "or a list."
                )

        if self.cluster_number is None:
            ks = np.arange(2, 21)
            losses = []
            for k in ks:
                kmeans = KMeans(k, random_state=0, n_init="auto")
                labels = kmeans.fit(self.theta).predict(self.theta)
                losses.append(-silhouette_score(self.theta, labels))
            losses = np.array(losses)
            minimum_index = np.argmin(losses)
            self.cluster_number = ks[minimum_index]

            kmeans = KMeans(self.cluster_number, random_state=0, n_init="auto")
            self.cluster_labels = kmeans.fit(self.theta).predict(self.theta)
            self.custom_cluster = False
        else:
            self.custom_cluster = True

        if self.cluster_number == 20:
            warnings.warn(
                "The number of clusters is 20. "
                + "This is the maximum "
                + "number of clusters that can be used. If you "
                + "require more clusters, please specify the "
                + "'cluster_number' kwarg. "
                + "margarine will continue "
                + "with 20 clusters."
            )

        if np.array(list(self.cluster_labels)).dtype == "float":
            # convert cluster labels to integers
            self.cluster_labels = self.cluster_labels.astype(int)
        # count the number of times a cluster label appears in cluster_labels
        self.cluster_count = np.bincount(self.cluster_labels)

        if self.custom_cluster:
            if self.cluster_count.min() < 100:
                warnings.warn(
                    "One or more clusters are too small "
                    + "(n_cluster < 100). "
                    + "Since cluster_number was supplied margarine"
                    + "will continue but may crash."
                )
        else:
            # While loop to make sure clusters are not too small
            while self.cluster_count.min() < 100:
                warnings.warn(
                    "One or more clusters are too small "
                    + "(n_cluster < 100). "
                    + "Reducing the number of clusters by 1."
                )
                minimum_index -= 1
                self.cluster_number = ks[minimum_index]
                kmeans = KMeans(
                    self.cluster_number, random_state=0, n_init="auto"
                )
                self.cluster_labels = kmeans.fit(self.theta).predict(
                    self.theta
                )
                self.cluster_count = np.bincount(self.cluster_labels)
                if self.cluster_number == 2:
                    # break if two clusters
                    warnings.warn(
                        "The number of clusters is 2. This is the "
                        + "minimum number of clusters that can be used. "
                        + "Some clusters may be too small and the "
                        + "train/test split may fail."
                        + "Try running without clusting. "
                    )
                    break

        split_theta = []
        split_sample_weights = []
        for i in range(self.cluster_number):
            split_theta.append(self.theta[self.cluster_labels == i])
            split_sample_weights.append(
                self.sample_weights[self.cluster_labels == i]
            )

        self.split_theta = split_theta
        self.split_sample_weights = split_sample_weights

        self.flow = []
        for i in range(len(split_theta)):
            self.flow.append(
                MAF(
                    split_theta[i],
                    weights=split_sample_weights[i],
                    number_networks=self.number_networks,
                    learning_rate=self.learning_rate,
                    hidden_layers=self.hidden_layers,
                    theta_min=self.theta_min,
                    theta_max=self.theta_max,
                    activation_func=self.activation_func,
                )
            )

    def train(
        self,
        epochs: int = 100,
        patience: int | None = None,
        early_stop: bool = False,
        loss_type: str = "sum",
    ) -> None:
        r"""Train the clusterMAF.

        This function is called to train the clusterMAF once it has been
        initialised. It calls the `train()` function for each flow.

        Example:
            >>> from margarine.clustered import clusterMAF
            >>> bij = clusterMAF(theta, weights=weights)
            >>> bij.train()

        Keyword Args:
            epochs (int, optional): The number of iterations to
                train the neural networks for. Defaults to 100.
            patience (int | None, optional): The number of epochs
                with no improvement
                on the test loss before early stopping is triggered.
                Defaults to 2% of epochs.
            early_stop (bool, optional): Whether to implement
                early stopping or train for the set number of epochs.
                If True, training stops when test loss
                has not improved for 2% of the requested epochs,
                and the model rolls back to the best performing version.
                Defaults to False.
            loss_type (str, optional): Whether to use the 'sum' or
                'mean' of the weighted log probabilities to
                calculate the loss function.
                Defaults to 'sum'.
        """
        for i in range(len(self.flow)):
            self.flow[i].train(
                epochs=epochs,
                patience=patience,
                early_stop=early_stop,
                loss_type=loss_type,
            )

    def log_prob(self, params: np.ndarray) -> np.ndarray:
        """Log-probability for a given set of parameters.

        While each density estimator has its own built in log probability
        function, a correction has to be applied for the transformation of
        variables that is used to improve accuracy when learning and we
        have to sum probabilities over the series of flows. The
        correction and the sum are implemented here.

        Args:
            params (tf.Tensor | np.ndarray): The set of samples for which to
                calculate the log probability.

        Returns:
            tf.Tensor: The log-probabilities of the provided samples.
        """
        # currently working with numpy not tensorflow
        # nan repalacement with 0 difficult in tensorflow

        flow_weights = [
            np.sum(weights) for weights in self.split_sample_weights
        ]
        flow_weights = np.array(flow_weights)
        flow_weights = flow_weights / np.sum(flow_weights)

        logprob = []
        for flow, weight in zip(self.flow, flow_weights):
            flow_prob = flow.log_prob(
                tf.convert_to_tensor(params, dtype=tf.float32)
            ).numpy()
            probs = flow_prob + np.log(weight)
            logprob.append(probs)
        logprob = np.array(logprob)
        logprob = logsumexp(logprob, axis=0)

        return logprob

    def log_like(
        self,
        params: tf.Tensor | np.ndarray,
        logevidence: float,
        prior_de: tfd.TransformedDistribution | None = None,
    ) -> tf.Tensor:
        r"""Return the log-likelihood for a given set of parameters.

        It requires the logevidence from the original nested sampling run
        in order to do this and in the case that the prior is non-uniform
        a trained prior density estimator should be provided.

        Args:
            params (tf.Tensor | np.ndarray): The set of samples for which to
                calculate the log-likelihood.
            logevidence (float): The log-evidence from the original
                nested sampling run.
            prior_de (tfd.TransformedDistribution, optional): A trained
                density estimator for the prior. If the prior is uniform,
                this can be left as None. Defaults to None.

        Returns:
            tf.Tensor: The log-likelihoods of the provided samples.
        """
        if prior_de is None:
            warnings.warn("Assuming prior is uniform!")

            n = (np.sum(self.sample_weights) ** 2) / (
                np.sum(self.sample_weights**2)
            )

            theta_max = np.max(self.theta, axis=0)
            theta_min = np.min(self.theta, axis=0)
            a = ((n - 2) * theta_max - theta_min) / (n - 3)
            b = ((n - 2) * theta_min - theta_max) / (n - 3)
            prior_logprob = np.log(
                np.prod([1 / (a - b) for i in range(len(b))])
            )
        else:
            prior_logprob = prior_de.log_prob(params).numpy()

        posterior_logprob = self.log_prob(params)

        loglike = posterior_logprob + prior_logprob - logevidence

        return loglike

    # @tf.function(jit_compile=True)
    def __call__(self, u: np.ndarray, seed: int = 1420) -> tf.Tensor:
        r"""Transform samples from the unit hypercube to the clusterMAF.

        This function is used when calling the clusterMAF class to transform
        samples from the unit hypercube to samples on the clusterMAF.

        Args:
            u (np.ndarray): Samples on the unit hypercube.
            seed (int, optional): Set the seed for the cluster choice.

        Returns:
            tf.Tensor: Samples transformed to the clusterMAF.
        """
        flow_weights = [
            np.sum(weights) for weights in self.split_sample_weights
        ]
        flow_weights = np.array(flow_weights)
        probabilities = flow_weights / np.sum(flow_weights)
        options = np.arange(0, self.cluster_number)

        np.random.seed(int(round(u[0][0] * 1000)))
        choice = np.random.choice(options, p=probabilities, size=len(u))
        np.random.seed(None)

        totals = [
            len(choice[choice == options[i]]) for i in range(len(options))
        ]
        totals = np.hstack([0, np.cumsum(totals)])

        values = []
        for i in range(len(options)):
            x = self.flow[i](u[totals[i] : totals[i + 1]])
            values.append(x)

        x = tf.concat(values, axis=0)
        return x

    def sample(self, length: int = 1000) -> tf.Tensor:
        r"""Generate samples on the clusterMAF.

        Args:
            length (int, optional): The number of samples to generate.
                Defaults to 1000.

        Returns:
            tf.Tensor: Samples generated on the clusterMAF.
        """
        u = np.random.uniform(size=(length, self.theta.shape[-1]))
        return self(u)

    def save(self, filename: str) -> None:
        r"""Save a trained clusterMAF.

        Args:
            filename (str): Path in which to save the pickled clusterMAF.
        """
        nn_weights = {}
        for i in range(self.cluster_number):
            made = self.flow[i].mades
            w = [m.get_weights() for m in made]
            nn_weights[i] = w

        with open(filename, "wb") as f:
            pickle.dump(
                [
                    self.theta,
                    nn_weights,
                    self.sample_weights,
                    self.number_networks,
                    self.hidden_layers,
                    self.learning_rate,
                    self.cluster_number,
                    self.cluster_labels,
                ],
                f,
            )

    @classmethod
    def load(cls, filename: str) -> "clusterMAF":
        r"""Load a saved clusterMAF from a file.

        Example:
            >>> from margarine.clustered import clusterMAF
            >>> file = 'path/to/pickled/MAF.pkl'
            >>> bij = clusterMAF.load(file)

        Args:
            filename (str): Path to the saved clusterMAF file.

        Returns:
            clusterMAF: The loaded clusterMAF object.
        """
        with open(filename, "rb") as f:
            data = pickle.load(f)
            (
                theta,
                nn_weights,
                sample_weights,
                number_networks,
                hidden_layers,
                learning_rate,
                cluster_number,
                labels,
            ) = data

        maf = cls(
            theta,
            weights=sample_weights,
            number_networks=number_networks,
            hidden_layers=hidden_layers,
            learning_rate=learning_rate,
            cluster_number=cluster_number,
            cluster_labels=labels,
        )
        maf(np.random.uniform(size=(1, theta.shape[-1])))
        for j in range(len(maf.flow)):
            for made, nnw in zip(maf.flow[j].mades, nn_weights[j]):
                made.set_weights(nnw)

        return maf



"""Tests for importance sampling functionality."""

import numpy as np
import pytest
import tensorflow as tf
from anesthetic import MCMCSamples
from scipy.stats import norm

from margarine.maf import MAF
from margarine.marginal_stats import calculate


def D_KL(
    logL: np.ndarray | tf.Tensor, weights: np.ndarray | tf.Tensor
) -> float:
    """Calculate the Kullback-Leibler divergence."""
    return -np.average(logL, weights=weights)


def d_g(
    logL: np.ndarray | tf.Tensor, weights: np.ndarray | tf.Tensor
) -> float:
    """Calculate the BMD statistic."""
    return 2 * np.cov(logL, aweights=weights)


norm = norm(loc=4.2, scale=0.3)
theta = norm.rvs(size=(1000, 2))
logL = norm.logpdf(theta).sum(axis=1)
weights = np.ones(len(theta))

mcmc_samples = MCMCSamples(data=theta, logL=logL)
samples_kl = D_KL(logL, weights)
samples_d = d_g(logL, weights)
names = [i for i in range(theta.shape[-1])]


class TestImportance:
    """Class to test importance sampling functionality."""

    @pytest.fixture
    def maf(self) -> MAF:
        """Fixture to create MAF instance."""
        return MAF(mcmc_samples, parameters=names)

    @pytest.fixture
    def calc(self, maf: MAF) -> calculate:
        """Fixture to create a calculate instance with the trained MAF."""
        return calculate(maf)

    def likelihood(self, parameters: np.ndarray) -> np.ndarray:
        """Define a simple Gaussian likelihood function."""
        y = np.array([0, 6])
        loglikelihood = (
            -0.5 * np.log(2 * np.pi * (1**2))
            - 0.5 * ((y - parameters) / 1) ** 2
        ).sum(axis=-1)
        return loglikelihood

    def prior(self, parameters: np.ndarray) -> np.ndarray:
        """Define a simple Gaussian prior function."""
        y = np.array([6, 0])
        density = (
            -0.5 * np.log(2 * np.pi * (1**2))
            - 0.5 * ((y - parameters) / 1) ** 2
        ).sum(axis=-1)
        return density

    # @pytest.mark.parametrize(("prior", [prior, maf.log_prob]))
    def test_integrate_prior(self, calc: calculate) -> None:
        """Test integration using the prior distribution."""
        calc.integrate(self.likelihood, self.prior, sample_size=1000)

    def test_integrate_maf(self, calc: calculate, maf: MAF) -> None:
        """Test integration using the MAF distribution."""
        calc.integrate(self.likelihood, maf.log_prob, sample_size=1000)

    def test_integrate_unphysical(self, calc: calculate) -> None:
        """Test integration with unphysical logzero value."""
        with pytest.raises(ValueError):
            calc.integrate(
                self.likelihood, self.prior, sample_size=1000, logzero=10
            )

    @pytest.mark.parametrize("ss", [10, 100])
    @pytest.mark.parametrize("bs", [10, 100])
    def test_integrate_batching(
        self, calc: calculate, ss: int, bs: int
    ) -> None:
        """Test integration with different sample and batch sizes."""
        calc.integrate(
            self.likelihood, self.prior, sample_size=ss, batch_size=bs
        )



"""The KDE class used to generate and work with Kernel Density Estimators."""

import pickle
import warnings

import anesthetic
import numpy as np
import tensorflow as tf
from anesthetic.samples import Samples
from scipy.optimize import root_scalar
from scipy.stats import gaussian_kde, norm
from tensorflow_probability import bijectors as tfb

import margarine
from margarine.processing import _forward_transform, _inverse_transform


class KDE:
    r"""Class used to generate and work with Kernel Density Estimators."""

    def __init__(
        self,
        theta: np.ndarray | Samples,
        **kwargs: dict,
    ) -> None:
        r"""Kernel Density Estimation (KDE) class for weighted samples.

        This class generates a KDE from weighted samples,
        generates new samples from
        the KDE, transforms samples from the hypercube to the
        KDE distribution, and provides methods to save and load KDE models.

        Args:
            theta (np.ndarray | Samples):
                The samples from the probability
                distribution to learn.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            weights (np.ndarray, optional): The weights associated
                with the samples. If an anesthetic NestedSamples or MCMCSamples
                object is passed, the weights are drawn from it.
                Defaults to np.ones(len(theta)).
            bw_method (str | float | callable, optional): The
                bandwidth method for the KDE. Can be a string
                ('scott', 'silverman'), scalar, or callable.
            theta_max (np.ndarray, optional): The true upper
                limits of the priors used to generate the samples.
            theta_min (np.ndarray, optional): The true lower
                limits of the priors used to generate the samples.
            parameters (list of str, optional): The relevant
                parameters to use. Only needed if theta is an
                anesthetic samples object. If not provided, all
                parameters will be used.

        Attributes:
            kde (scipy.stats.gaussian_kde): The generated KDE object.
                Available after calling `generate_kde()`.
                Initialization and KDE generation are kept
                separate to allow effective model saving and loading.
            theta_max (np.ndarray): The true upper limits of the priors. If not
                supplied as a keyword argument, this is an
                approximate estimate.
            theta_min (np.ndarray): The true lower limits of the priors. If not
                supplied as a keyword argument, this is an
                approximate estimate.

        Example:
            >>> from margarine.kde import KDE
            >>> import numpy as np
            >>> theta = np.loadtxt('path/to/samples.txt')
            >>> weights = np.loadtxt('path/to/weights.txt')
            >>> kde_model = KDE(theta, weights=weights)
            >>> kde_model.generate_kde()
            >>> # Access the KDE via kde_model.kde
        """
        self.parameters = kwargs.pop("parameters", None)

        if isinstance(
            theta,
            anesthetic.samples.NestedSamples | anesthetic.samples.MCMCSamples,
        ):
            weights = theta.get_weights()
            if self.parameters:
                theta = theta[self.parameters].values
            else:
                if isinstance(theta, anesthetic.samples.NestedSamples):
                    self.parameters = theta.columns[:-3].values
                    theta = theta[theta.columns[:-3]].values
                else:
                    self.parameters = theta.columns[:-1].values
                    theta = theta[theta.columns[:-1]].values
        else:
            weights = kwargs.pop("weights", np.ones(len(theta)))

        self.theta = theta
        self.sample_weights = weights

        self.n = (np.sum(weights) ** 2) / (np.sum(weights**2))
        theta_max = np.max(theta, axis=0)
        theta_min = np.min(theta, axis=0)
        a = ((self.n - 2) * theta_max - theta_min) / (self.n - 3)
        b = ((self.n - 2) * theta_min - theta_max) / (self.n - 3)
        self.theta_min = kwargs.pop("theta_min", b)
        self.theta_max = kwargs.pop("theta_max", a)

        self.bw_method = kwargs.pop("bw_method", "silverman")

    def generate_kde(self) -> gaussian_kde:
        r"""Generates a weighted KDE."""
        theta = tf.convert_to_tensor(self.theta, dtype=tf.float32)
        theta_min = tf.convert_to_tensor(self.theta_min, dtype=tf.float32)
        theta_max = tf.convert_to_tensor(self.theta_max, dtype=tf.float32)

        phi = _forward_transform(theta, theta_min, theta_max).numpy()
        mask = np.isfinite(phi).all(axis=-1)
        phi = phi[mask, :]
        weights_phi = self.sample_weights[mask]
        weights_phi /= weights_phi.sum()

        self.kde = gaussian_kde(
            phi.T, weights=weights_phi, bw_method=self.bw_method
        )

        return self.kde

    def __call__(self, u: np.ndarray) -> np.ndarray:
        r"""Transform samples from the unit hypercube to samples on the KDE.

        Args:
            u (np.ndarray): Samples from the unit hypercube
                to be transformed.

        Returns:
            np.ndarray: The transformed samples.
        """
        # generate useful parameters for __call__ function to transform
        # hypercube into samples on the KDE.
        S = self.kde.covariance
        mu = self.kde.dataset.T
        steps, s = [], []
        for i in range(mu.shape[-1]):
            step = S[i, :i] @ np.linalg.inv(S[:i, :i])
            steps.append(step)
            s.append((S[i, i] - step @ S[:i, i]) ** 0.5)

        # transform samples from unit hypercube to kde
        transformed_samples = []
        for j in range(len(u)):
            x = u[j]
            y = np.zeros_like(x)
            for i in range(len(x)):
                m = mu[:, i] + steps[i] @ (y[:i] - mu[:, :i]).T
                y[i] = root_scalar(
                    lambda f: (
                        norm().cdf((f - m) / s[i]) * self.kde.weights
                    ).sum()
                    - x[i],
                    bracket=(mu[:, i].min() * 2, mu[:, i].max() * 2),
                    method="bisect",
                ).root
            transformed_samples.append(
                _inverse_transform(
                    tf.convert_to_tensor(y, dtype=tf.float32),
                    tf.convert_to_tensor(self.theta_min, dtype=tf.float32),
                    tf.convert_to_tensor(self.theta_max, dtype=tf.float32),
                )
            )
        transformed_samples = np.array(transformed_samples)
        return transformed_samples

    def sample(self, length: int = 1000) -> tf.Tensor:
        r"""Direct sampling from the KDE.

        Function can be used to generate samples from the KDE. It is much
        faster than the __call__ function but does not transform samples
        from the hypercube onto the KDE. It is however useful if we
        want to generate a large number of samples that can then be
        used to calulate the marginal statistics.

        Args:
            length (int): The number of samples to generate from the KDE.

        Returns:
            tf.Tensor: Samples generated from the KDE.
        """
        x = self.kde.resample(length).T
        return _inverse_transform(
            tf.convert_to_tensor(x, dtype="float32"),
            tf.convert_to_tensor(self.theta_min, dtype=tf.float32),
            tf.convert_to_tensor(self.theta_max, dtype=tf.float32),
        )

    def log_prob(self, params: np.ndarray) -> np.ndarray:
        """Caluclate the log-probability for a given set of parameters.

        While the density estimator has its own built in log probability
        function, a correction has to be applied for the transformation of
        variables that is used to improve accuracy when learning. The
        correction is implemented here.

        Args:
            params (np.ndarray): The set of samples for which to calculate
                the log probability.

        Returns:
            np.ndarray: The log-probability for the given samples.
        """
        mins = self.theta_min.astype(np.float32)
        maxs = self.theta_max.astype(np.float32)

        transformed_x = _forward_transform(
            tf.convert_to_tensor(params, dtype=tf.float32),
            tf.convert_to_tensor(mins, dtype=tf.float32),
            tf.convert_to_tensor(maxs, dtype=tf.float32),
        ).numpy()

        transform_chain = tfb.Chain(
            [
                tfb.Invert(tfb.NormalCDF()),
                tfb.Scale(1 / (maxs - mins)),
                tfb.Shift(-mins),
            ]
        )

        def norm_jac(y: np.ndarray) -> np.ndarray:
            """Calculate the normalising jacobian for the transformation."""
            return transform_chain.inverse_log_det_jacobian(
                y, event_ndims=0
            ).numpy()

        correction = norm_jac(transformed_x)
        if params.ndim == 1:
            logprob = (
                self.kde.logpdf(transformed_x.T) - np.sum(correction)
            ).astype(np.float64)
        else:
            logprob = (
                self.kde.logpdf(transformed_x.T) - np.sum(correction, axis=1)
            ).astype(np.float64)

        return logprob

    def log_like(
        self,
        params: np.ndarray,
        logevidence: float,
        prior: np.ndarray | None = None,
        prior_weights: np.ndarray | None = None,
    ) -> np.ndarray:
        r"""Return the log-likelihood for a given set of parameters.

        It requires the logevidence from the original nested sampling run
        in order to do this and in the case that the prior is non-uniform
        prior samples should be provided.

        Args:
            params (np.ndarray): The set of samples for which to
                calculate the log probability.
            logevidence (float): Should be the log-evidence from
                the full nested sampling run with nuisance parameters.

            prior (np.ndarray | None): An array of prior samples
                corresponding to the prior. Default assumption is that
                the prior is uniform which is required if you want to
                combine likelihoods from different experiments/data sets.
                In this case samples and prior samples should be
                reweighted prior to any training.
            prior_weights (np.ndarray | None): The weights associated
                with the prior samples.

        Returns:
            np.ndarray: The log-likelihood for the given samples.
        """
        if prior is None:
            warnings.warn("Assuming prior is uniform!")
            prior_logprob = np.log(
                np.prod(
                    [
                        1 / (self.theta_max[i] - self.theta_min[i])
                        for i in range(len(self.theta_min))
                    ]
                )
            )
        else:
            self.prior = margarine.kde.KDE(prior, weights=prior_weights)
            self.prior.generate_kde()
            prior_logprob_func = self.prior.log_prob
            prior_logprob = prior_logprob_func(params)

        posterior_logprob = self.log_prob(params)

        loglike = posterior_logprob + logevidence - prior_logprob

        return loglike

    def save(self, filename: str) -> None:
        r"""Save an instance of the KDE class and associated generated KDE.

        Args:
            filename (str): Path to save the KDE to.
        """
        with open(filename, "wb") as f:
            pickle.dump([self.theta, self.sample_weights, self.kde], f)

    @classmethod
    def load(cls, filename: str) -> "KDE":
        r"""Load a saved KDE from a file.

        Example:
            >>> from margarine.kde import KDE
            >>> file = 'path/to/pickled/bijector.pkl'
            >>> kde_model = KDE.load(file)

        Args:
            filename (str): Path to the saved KDE file.

        Returns:
            KDE: The loaded KDE object.
        """
        with open(filename, "rb") as f:
            theta, sample_weights, kde = pickle.load(f)

        kde_class = cls(theta, weights=sample_weights)
        kde_class.kde = kde
        kde_class(np.random.uniform(0, 1, size=(2, theta.shape[-1])))

        return kde_class
